void KING_RunFrame(fx_vdc_t **, uint32 *pXBuf, MDFN_Rect *LineWidths, int skip);
void KING_SetPixelFormat(int rshift, int gshift, int bshift);
uint32 KING_GetADPCMControl(void);
uint8 FXVCE_Read8(uint32 A);
void FXVCE_Write8(uint32 A, uint8 V);
uint16 KING_Read16(uint32 A);
uint32 KING_Read32(uint32 A);
void KING_Write16(uint32 A, uint16 V);
void KING_Write32(uint32 A, uint32 V);
bool KING_Init(void);
uint8 KING_GetADPCMNibble(int ch);

